package cologne.eck.peafactory.crypto.kdf;

import cologne.eck.all_peas.data.PeaProperties;
import cologne.eck.peafactory.crypto.CipherStuff;
import cologne.eck.peafactory.crypto.HashStuff;
import cologne.eck.tools.Comparator;
import cologne.eck.tools.TestLog;
import cologne.eck.tools.UnexpectedValueException;


/*
 * Peafactory - Production of Password Encryption Archives
 * Copyright (C) 2015  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

/**
 * parent class of all used key derivation functions 
 */

public abstract class KeyDerivation {
	
	private static KeyDerivation kdf = null;
	
/*	private static int tCost = 0;
	private static int mCost = 0;
	private static int arg3 = 0;
	private static int arg4 = 0;
	private static int arg5 = 0;
	private static int arg6 = 0;*/
	
	private static int saltSize = 32;// at most 129 byte for fixed initialized salt
	private static byte[] salt;
	

	private static byte[] extraValues = "peafactory_extra".getBytes(); // 16 byte
	
	private static String versionString = "";
	
	/**
	 * Derive the key. All parameters are set in child classes
	 * 
	 * @param pswMaterial	material build from the password
	 * 
	 * @return				the derived key
	 */
	public abstract byte[] deriveKey( byte[] pswMaterial);
	
	/**
	 * Get the name of the key derivation scheme
	 * 
	 * @return the name of the key derivation scheme
	 */
	public abstract String getName();
	
	/**
	 * Get the required memory for the key derivation function. 
	 * Returns the required memory in MiB. 
	 * 
	 * @return	memory requirement in MiB
	 */
	public abstract int getMemoryRequirement();

	/**
	 * Get a key of required length (expand or extract) using HKDF; 
	 * this is only used for Bcrypt and Scrypt
	 * 
	 * @param keyMaterial	the material from key derivation
	 * @return				the key
	 */
	protected final static byte[] adjustKeyMaterial(byte[] keyMaterial){
		
		if (PeaProperties.isTestMode()){
			Comparator.checkNullVector(keyMaterial);
		}
		//Help.printBytes("key", HashStuff.generateHKDFBytes(keyMaterial, null, CipherStuff.getKeySize()));
		
		return HashStuff.generateHKDFBytes(keyMaterial, salt, CipherStuff.getKeySize());
	}
	
	/**
	 * Get info (name, version, parameters) about the used KDF
	 * 
	 * @return info about the used KDF and parameters as String
	 */
	public abstract String getInfoString();
	
	/**
	 * Print informations about the used key derivation function
	 * 
	 * @param print 	if true, informations are printed
	 */
/*	public final static void printInfos(boolean print) {
		
		
		if (print == true) {
			System.out.print("Key derivation: " +  kdf.getName() );
			
			if (! versionString.equals("")){
				System.out.print(", " + versionString);
			}
			System.out.print(", time parameter: " + tCost 
					+ ", memory parameter: " + mCost
					);
			if (arg3 != 0) {
				System.out.print(", parameter 3: " + arg3);
			}
			if (arg4 != 0) {
				System.out.print(", parameter 4: " + arg4);
			}
			if (arg5 != 0) {
				System.out.print(", parameter 5: " + arg5);
			}
			if (arg6 != 0) {
				System.out.print(", parameter 6: " + arg6);
			}
			TestLog.o(KeyDerivation.class, "");
		}		
	}*/

	//=============================================
	// Getter & Setter  
	
	/**
	 * Get the String for the special instance if there is one.
	 * "i", "id", "d" for Argon2, 
	 * "Dragonfly", "Dragonfly-Full", "Butterfly", "Butterfly-Full" for Catena, 
	 * null for Pomelo, Scrypt, Bcrypt
	 * 
	 * @return the kdfInstance
	 */
/*	public static String XgetKdfInstanceAppendix() {
		return kdfInstance;
	}*/

	/**
	 * et the String for the special instance if there is one.
	 * "i", "id", "d" for Argon2, 
	 * "Dragonfly", "Dragonfly-Full", "Butterfly", "Butterfly-Full" for Catena, 
	 * null for Pomelo, Scrypt, Bcrypt
	 * 
	 * @param kdfInstance the kdfInstance to set
	 */
/*	public static void XsetKdfInstanceAppendix(String kdfInstance) {
		KeyDerivation.kdfInstance = kdfInstance;
	}*/

	/**
	 * Set the parameters for the key derivation function: 
	 * index 0 time cost, index 1 memory cost, index 2 parallelism
	 * 
	 * @param params	index 0 time cost, index 1 memory cost, index 2 parallelism
	 */
/*	public static final void setParameters(int...params) {
		if (params == null ) { // at least time parameter
			new UnexpectedValueException("byte[]", "params", "is null").printDescription();
			PeaProperties.getVmBridge().exitWithError(null, 
					PeaProperties.getVmBridge().translate("unexpected_error") + " Missing parameters for KDF", false);
			return;
		}
		int paramLen = params.length;
		if (paramLen < 1) { // at least time parameter
			new UnexpectedValueException("byte[]", "params", "is empty").printDescription();
			PeaProperties.getVmBridge().exitWithError(null, 
					PeaProperties.getVmBridge().translate("unexpected_error") + " Empty parameters for KDF", false);
		}
		// set the parameters:
		for (int i = 0; i < paramLen; i++) {
			if (i == 0) {
				tCost = params[0];
			} else if (i == 1) {
				mCost = params[1];
			} else if (i == 2){
				arg3 = params[2];
			} else if (i == 3) {
				arg4 = params[3];
			} else if (i == 4) {
				arg5 = params[4];
			} else if (i == 5) {
				arg5 = params[5];
			} else {
				new UnexpectedValueException("int", "param[i]", "invalid parameter: " + params[i]).printDescription();;
			}
		}
		
	}*/
	
	
	/**
	 * @return the tCost
	 */
	public abstract int gettCost();

	/**
	 * @param _tCost the tCost to set
	 */
	/*public static final void settCost(int _tCost) {
		// TODO Auto-generated method stub
		
	}*/

	/**
	 * @return the mCost
	 */
	public abstract int getmCost();

	/**
	 * @param _mCost the mCost to set
	 */
	//public abstract void setmCost(int _mCost);

	/**
	 * @return the arg3
	 */
	public abstract int getArg3();

	/**
	 * @param _arg3 the arg3 to set
	 */
	//public abstract void setArg3(int _arg3);

	/**
	 * @return the arg4
	 */
	public abstract int getArg4();

	/**
	 * @param _arg4 the arg4 to set
	 */
	//public abstract void setArg4(int _arg4);

	/**
	 * @return the arg5
	 */
	public abstract int getArg5();

	/**
	 * @param _arg5 the arg5 to set
	 */
	//public abstract void setArg5(int _arg5);

	/**
	 * @return the arg6
	 */
	public abstract int getArg6();

	/**
	 * @param _arg6 the arg6 to set
	 */
	//public abstract void setArg6(int _arg6);

	/**
	 * @return the salt
	 */
	public static byte[] getSalt() {
		if (salt == null) {
			TestLog.v(KeyDerivation.class, "Accessed salt is yet null");
			//estLog.ve(KeyDerivation.class, "Salt is null", 5);
		}
		return salt;
	}

	/**
	 * @param _salt the salt to set
	 */
	public static void setSalt(byte[] _salt) {		
		//TestLog.ve(KeyDerivation.class, "set salt "+Converter.bytes2hex(_salt), false, 3);
		if (_salt == null) {
			salt = null;
			new UnexpectedValueException("byte[]","salt","set salt is null").printDescription();			
		} else {
			if (_salt.length < saltSize) {
				if (_salt.length < 16) {
					new UnexpectedValueException("int","_salt.length","is too short: " + _salt.length).printDescription();			
				} else {
					TestLog.e(KeyDerivation.class, "Warning: short salt size: " + _salt.length);
				}
			}
			Comparator.checkNullVector(_salt);
			salt = new byte[_salt.length];
			System.arraycopy(_salt,  0,  KeyDerivation.salt,  0,  _salt.length);
		}
//TestLog.v(KeyDerivation.class, "****set salt", 5);
		//Help.printBytes("KeyDerivation set Salt ", salt);
//new UnexpectedValueException("","SALT",new String(_salt)).printDescription();
	} 

	/**
	 * @return the extraValues
	 */
	protected static byte[] getExtraValues() {
		return extraValues;
	}

	/**
	 * @param _extraValues the extraValues to set
	 */
/*	public static void setExtraValues(byte[] _extraValues) {
		KeyDerivation.extraValues = _extraValues;
	}*/

	/**
	 * @return the kdf
	 */
	public static KeyDerivation getKdf() {
		return kdf;
	}

	/**
	 * Set the key derivation function.
	 * The KDF should always be set with 
	 * setting the parameters in the constructor
	 * 
	 * @param kdf the kdf to set
	 */
	public static void setKdf(KeyDerivation kdf) {
//TestLog.o(getClass(), "Set " + kdf.getInfoString());
		KeyDerivation.kdf = kdf;
		//settCost(kdf.gettCost());
	}

	/**
	 * @return the versionString
	 */
	public static String getVersionString() {
		return versionString;
	} 

	/**
	 * @param versionString the versionString to set
	 */
	public static void setVersionString(String versionString) {
		KeyDerivation.versionString = versionString;
	}
	
	public static int getSaltSize(){
		if (salt == null) {
			return saltSize;
		} else {
			return salt.length;
		}
	}
	/**
	 * This is for testing only
	 * 
	 * @param newSaltSize
	 */
	protected static void setSaltSize(int newSaltSize) {
		TestLog.e(KeyDerivation.class, "KeyDerivation: set salt size for testing: " + newSaltSize);
		saltSize = newSaltSize;
	}
}
